from datetime import datetime


class Order(object):
    def __init__(
        self,
        order_id,   # type: int
        date,       # type: datetime
        state,      # type: int
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.date = date
        self.state = state
