# -*- coding: utf-8 -*-


class RuptureEvents(object):
    FullRuptureRequest = "FullRuptureRequest"
    CleanRupture = "CleanRupture"
    OrderModified = "ORDER_MODIFIED"


class RuptureEventTypes(object):
    CleanRuptureThread = "CleanRuptureThread"
    CleanRuptureOnOpenDay = "OnOpenDay"
