from typing import (
    Dict,
    List,
    Tuple,
)

from ._DeviceConfig import DeviceConfig


class Configs(object):

    def __init__(
        self,
        component_name,     # type: str
        required_services,  # type: str
        persistence_name,   # type: str
        sector_config,      # type: Dict[str, Dict[str, Tuple[str, Dict[int, int]]]]
        device_configs,     # type: List[DeviceConfig]
        brand_config,       # type: Dict[str, Dict[str, str]]
    ):
        # type: (...) -> None

        self.component_name = component_name
        self.required_services = required_services
        self.persistence_name = persistence_name
        self.sector_config = sector_config
        self.device_configs = device_configs
        self.brand_config = brand_config
